'use strict';

const cluster = require('cluster');
const events = require('events');

function ClusterManager() {
    if (!cluster.isMaster) {
        throw new Error('ClusterManager must live on the master process');
    }

    events.EventEmitter.call(this);

    this._boundOnMessage = this._onMessage.bind(this);

    this.numWorkers = 0;
    this.verbose = false;
    this.spawning = false;
    this.ready = false;

    this.job = {
        name: '',
        running: false,
        results: [],
        callback: null
    };
}

ClusterManager.prototype = Object.create(events.EventEmitter.prototype);

Object.assign(ClusterManager.prototype, {
    /**
     * Spawns all the workers for this manager to track.
     *
     * @method
     * @param {number} numWorkers - The number of workers to spawn.
     * @param {string|object} [settings] - The path for the file to spawn for workers,
     *  or the settings object to pass into `cluster.setupMaster()` if any.
     * @param {function} [callback] - The callback to call when the workers are ready.
     */
    spawn: function (numWorkers, settings, callback) {
        if (this.ready || this.spawning)
            return this._error('Cannot call spawn when spawning or after workers are already spawned.');

        if (typeof settings === 'function') {
            callback = settings;
            settings = null;
        }
        else if (typeof settings === 'string') {
            settings = { exec: settings };
        }

        this.numWorkers = numWorkers;
        this.spawning = true;

        cluster.setupMaster(settings);

        this._log(`Spawning ${this.numWorkers} workers.`);

        // Setup event listeners
        cluster.on('exit', (worker, code, signal) => {
            this._error(worker, `Worker (id: ${worker.id}, pid: ${worker.process.pid}) died.`);
        });

        let onlineWorkers = 0;
        cluster.on('online', (worker) => {
            if (++onlineWorkers === this.numWorkers) {
                this._log(`${this.numWorkers} workers are online.`);
                if (callback) callback();
                this._ready();
            }
        });

        // fork the workers
        for (let i = 0; i < this.numWorkers; ++i) {
            const worker = cluster.fork();

            worker.on('message', (msg) => {
                this._onMessage(worker, msg);
            });
        }
    },

    dispatchJob: function (jobName, data, callback) {
        if (this.job.running)
            return this._error('Cannot dispatch new job while a job is already in progress.');

        this.job.name = jobName;
        this.job.running = true;
        this.job.results = [];
        this.job.callback = callback;

        // split data into each worker, send job to them
        const batchSize = Math.floor(data.length / this.numWorkers);
        const remainder = data.length % this.numWorkers;

        let count = 0;

        Object.keys(cluster.workers).forEach((id, i) => {
            const endIndex = count + batchSize + (i === this.numWorkers - 1 ? remainder : 0);
            const worker = cluster.workers[id];

            worker.send({
                job: jobName,
                data: data.slice(count, endIndex)
            });

            count = endIndex;
        });
    },

    _onMessage: function (worker, payload) {
        this.job.results.push(payload.results);

        if (this.job.results.length === this.numWorkers) {
            flatten(this.job.results);
            this.job.running = false;
            this._complete();
        }
    },

    _log: function () {
        if (this.verbose) {
            console.log.apply(console, arguments);
        }
    },

    _error: function (worker, msg) {
        this.emit('error', new Error(msg), worker);
    },

    _ready: function () {
        this.ready = true;
        this.emit('ready');
    },

    _complete: function () {
        this.job.callback(this.job);
        this.emit('complete', this.job);
    }
});

module.exports = new ClusterManager();

// https://jsperf.com/flatten-array-in-recursion-and-while-loop/7
function flatten(arr) {
    for (var i = 0; i < arr.length; ++i) {
        if (Array.isArray(arr[i])) {
            // need to add i and 1 as arguments to the real splice command,
            // plus all the values of arr[i]
            arr[i].splice(0, 0, i, 1);
            Array.prototype.splice.apply(arr, arr[i]);

            // Check for more inner arrays at beginning of this array
            --i;
        }
    }
}
